(function() {
  "use strict";
  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    FriendshipProtocol: {
      process: function(receivedData, storedData, senderId) {
        var handler;

        if (_(storedData).isUndefined() && receivedData.type !== 0) {
          throw new Error("Unexpected message of type \"" +
              _(this.MessageType).keyOf(receivedData.type) +
              "\" while \"" +
              _(this.MessageType).keyOf(0) +
              "\" was expected.");
        }

        if (_(storedData).has("expectation") &&
            receivedData.type !== storedData.expectation) {
          throw new Error("Unexpected message of type \"" +
              _(this.MessageType).keyOf(receivedData.type) +
              "\" while \"" +
              _(this.MessageType).keyOf(storedData.expectation) +
              "\" was expected.");
        }

        if (_(this.MessageHandler[receivedData.type]).isUndefined()) {
          throw new Error("Unknown message type: " + receivedData.type);
        }

        handler = this[this.MessageHandler[receivedData.type]];

        return handler.call(this, receivedData, storedData, senderId);
      },
      /**
       * Creates an instance of the input FriendshipProtocols, invokes them and
       * returns a Promise which will resolve when all of them completed their
       * operations.
       *
       * @param {FriendshipProtocol[]} handlers an array of
       *        FriendshipProtocol objects to handle.
       * @param {Snake.Model.Message} message the message to populate with the
       *        information collected from the FriendshipProtocols.
       * @param {Friend} friendship reference to the private descriptor of the
       *        friend we're dealing with.
       * @param {Number} senderId the identifier of the other party.
       * @returns {Promise} returns a Promise which will be resolve once all
       *          the operations that the handlers have to perform are complete.
       *          These promises will return an object with the information that
       *          have to be saved privately ("store" field) and those that have
       *          have to be sent to the other party ("send" field).
       * @static
       */
      handleProtocols: function(handlers, message, friendship, senderId) {
        senderId = Snake.identityPromise(senderId);

        return _(handlers).map(function(handler) {
          var fieldName, toProcess;

          fieldName = handler.fieldName;

          if (!message.has(fieldName)) return;

          toProcess = _({id: message.id}).extend(message.get(fieldName));
          return handler.process(toProcess, friendship[fieldName],
              senderId);
        });
      }
    }
  });
}());
