(function() {
  "use strict";

  var Crypto = Snake.Crypto;
  var OfflineAuthenticator = Crypto.OfflineAuthenticator;

  _.extend(Crypto, {}, {
    /**
     * Public key authenticator manually asking the user if the public keys is
     * correct.
     *
     * @class Snake.Crypto.ManualAuthenticator
     * @augments Snake.Crypto.OfflineAuthenticator
     */
    ManualAuthenticator: _(_(OfflineAuthenticator).clone()).extend({
      name: "ManualAuthenticator",
      verify: function(id) {
        var newFriend;

        // Start by fetching other user's details
        newFriend = new Snake.Model.User({id: id}, {});

        return newFriend.fetch().result.then(function() {
          return this.askUser(newFriend.get("username"),
                              util.abv2hex(newFriend.get("staticPublicKey")));
        }.bind(this)).then(function(answer) {
          if (answer) {
            return OfflineAuthenticator.AuthResult.Valid;
          }
          return OfflineAuthenticator.AuthResult.Unknown;
        }.bind(this));
      }
    })
  });
}());
