(function() {
  "use strict";

  var Crypto = Snake.Crypto;
  var Config = Snake.Config;

  _.extend(Crypto, {}, {
    /**
     * Base class for offline public key authenticators.
     * This type of authenticators do not require an interaction with the party
     * we're trying to authenticate.
     *
     * @class Snake.Crypto.OfflineAuthenticator
     */
    OfflineAuthenticator: {
      /**
      * Validity status of a public key.
      * @readonly
      * @memberOf Snake.Crypto.OfflineAuthenticator
      * @enum {Number}
      */
      AuthResult: {
        /**
         * It was not possible to determine whether the public key is valid or
         * not.
         * **/
        "Unknown": 0,
        /**
         * The public key is explicitly invalid (fake).
         */
        "Invalid": 1,
        /**
         * The public key is fully valid.
         */
        "Valid": 2
      },
      /**
       * Returns true if the OfflineAuthenticator was able to verify that the
       * public key associate with the specified user actually belongs to him.
       *
       * @param {Number} id identifier of the user we're trying to authenticate.
       * @returns {Snake.Crypto.OfflineAuthenticator.AuthenticationResult}
       *          the result of the authentication process for the user.
      *  @memberOf Snake.Crypto.OfflineAuthenticator
       * @abstract
       */
      verify: function(id) {
        throw new Error("Not implemented");
      },
      /**
       * Checks whether the current authenticator is enabled.
       *
       * @memberOf Snake.Crypto.OfflineAuthenticator
       * @returns {Boolean} returns true if this authenticator is enabled.
       */
      isEnabled: function() {
        return _.contains(Config.Crypto.offlineAuthenticators, this.name);
      }
    }
  });
}());
