(function() {
  "use strict";

  // TODO: remove me, use signatures
  function hashModel(data) {
    Snake.Logging.log("network", Snake.Logging.stringify(data));
    return Snake.Crypto.Common.hash(msgpack.pack(data));
  }

  // TODO: should these function be moved to Snake.Model.ModelBase?
  function foldData(data, schema, url) {
    var privateData = {}, format, noCryptFields;

    noCryptFields = schema.publicFields.concat(schema.authenticationOnlyFields);

    for (var key in data) {
      if (data.hasOwnProperty(key)) {
        format = schema.properties[key].format;

        // Check if it's a private field
        if (noCryptFields.indexOf(key) === -1) {
          privateData[key] = data[key];
          delete data[key];
        } else if (!_(format).isUndefined() &&
            Snake.Schema.binaryFormats.indexOf(format) !== -1) {
          // It's a public field, and it's a binary format we have to transform
          // it to Base128
          data[key] = Snake.IO.serializeBinary(data[key]);
        }
      }
    }

    data.object = url;

    // Pack with RJSON
    data["privateData"] = RJSON.pack(privateData);
    return data;
  }

  function getSchema(model) {
    if (!_.has(model, "urlRoot")) model = model.constructor;
    return Snake.Schema.getSchemaByUrlRoot(_.result(model, "urlRoot"));
  }

  function getOutgoingSchema(model) {
    if (!_.has(model, "urlRoot")) model = model.constructor;
      return Snake.OutgoingSchema.getSchemaByUrlRoot(_.result(model, "urlRoot"));
  }

  function signAndCrypt(data, model) {
    var dataToSign, iv, schema, noCryptFields, attributes;

    attributes = _.deepClone(data);

    data = foldData(data, getSchema(model),
      _.result(model, "urlRoot"));

    schema = getOutgoingSchema(model);
    noCryptFields = schema.publicFields.concat(
        schema.authenticationOnlyFields);

    // TODO: replace escaped chars
    // TODO: use UTF8
    dataToSign = msgpack.pack(attributes);

    // Sign first
    // bind takes two, parameters, we bind the first to the data to encrypt
    //  and the second comes from the result of getSigningKey
    return model.getSigningPrivateKey().then(function(signingKey) {
      // If a key is available sign
      if (!_(signingKey).isUndefined()) {
        return Snake.Crypto.Common.sign(dataToSign, signingKey);
      }
    }).then(function(signature) {

      // Wrap data with signature
      data.privateData = _({
        data: data.privateData,
        signature: signature
      }).filterUndefined();

      // Retrieve the encryption key and generate the IV in parallel
      return Promise.every(model.getEncryptionKey(),
          Snake.Crypto.Common.generateIV()).then(function(results) {
        var serialized, key, additionalData;

        key = results[0];
        serialized = msgpack.pack(data.privateData);

        // There's no key...
        if (_(key).isUndefined()) {
          if (data.isPublic === true ||
              _.isEqual(_(schema.publicFields).without("object"),
                        _(schema.properties).keys())) {
            // Makes sense: it's public
            // Note: iv will be undefined
            return serialized;
          }
          throw new Error("No encryption key available");
        }

        iv = results[1];
        // Serialize data with Message-pack
        additionalData = msgpack.pack(_(_(attributes)
            .pick(noCryptFields))
            .omit(["privateData", "date", "id", "object"]));
        Snake.Logging.log("network", "additionalData: " +
            Snake.Logging.stringify(_(_(attributes)
            .pick(noCryptFields))
            .omit(["privateData", "date", "id", "object"])));

        // Encrypt data
        return Snake.Crypto.Common.symmAuthEncryption(key, serialized, iv,
            additionalData);
      }.bind(this)).then(function(ciphertext) {
        var publicData;

        // Public data is composed of ciphertext, IV and...
        publicData = _({
          data: ciphertext,
          iv: iv
        }).filterUndefined();

        // ...all the authenticationOnlyFields
        _(publicData).extend(_(data).pick(schema.authenticationOnlyFields));
        // Which have to be removed from outside privateData
        data = _(data).omit(schema.authenticationOnlyFields);

        data.privateData = JSON.stringify(publicData, function(key, value) {
          // Base128 encode binary public fields
          if (value instanceof Uint8Array) {
            return Snake.IO.serializeBinary(value);
          }
          return value;
        });
        return data;
      });
    }).catch(Snake.errorHandler.bind("crypto"));
  }

  function decodeBinary(data, schema) {
    var format, binaryFields, recursiveSearch, recursiveDecode;

    // Base128-decode binary public fields
    for (var key in _(data).omit("privateData")) {
      if (data.hasOwnProperty(key)) {
        if (schema.properties[key].type === "object") {
          format = schema.properties[key].format;

          if (!_(format).isUndefined() &&
              Snake.Schema.binaryFormats.indexOf(format) !== -1) {
            data[key] = Snake.IO.deserializeBinary(data[key]);
          } else {

            // Collect the list of names of fields associated with binary data
            // TODO: this is not completly safe, as there might be two fields
            // with the same name but differente data types
            recursiveSearch = function(obj) {
              var props = obj;

              // Go through each field
              return _(_(props).keys()).reduce(function(prev, cur) {

                // Consider only objects
                if (_(props[cur]).isObject()) {
                  // Is that what we're looking for?
                  if (props[cur].type === "object" &&
                      Snake.Schema.binaryFormats.indexOf(
                      props[cur].format) !== -1) {
                    // Save it
                    return prev.concat([cur]);
                  } else {
                    // Recurse
                    return prev.concat(recursiveSearch(props[cur]));
                  }
                }
                // Nothing to do
                return prev;
              }, []);
            };

            binaryFields = recursiveSearch(schema.properties[key]);

            if (binaryFields.length > 0) {
              recursiveDecode = function(obj) {
                // Loop through all the keys
                _(_(obj).keys()).each(function(key) {
                  // Is a binary field?
                  if (_(obj[key]).isString() &&
                      binaryFields.indexOf(key) !== -1) {
                    // Decode it
                    obj[key] = Snake.IO.deserializeBinary(obj[key]);
                  } else if (_(obj[key]).isObject()) {
                    // If it's an object recurse
                    recursiveDecode(obj[key]);
                  }
                });
              };

              recursiveDecode(data[key]);
            }
          }
        }
      }
    }

    return data;
  }

  function decryptAndVerify(data, modelClass) {
    var iv, ciphertext, privateData, additionalData, schema, dataToVerify,
        findUnderscore;

    schema = getSchema(modelClass);
    if (!_(data).has("privateData")) {
      return Snake.identityPromise(decodeBinary(data, schema));
    }

    privateData = JSON.parse(data.privateData);
    delete data.privateData;

    // Move authentication only fields back with the public fields
    _(data).extend(_(privateData).pick(schema.authenticationOnlyFields));

    // Keep only the public fields, we don't want to have injections with
    // fields that should be private
    data = _(data).pick(schema.publicFields.concat(
        schema.authenticationOnlyFields));

    data = decodeBinary(data, schema);

    // Get the data that needs to be signed
    additionalData = msgpack.pack(
        _(data).omit(["privateData", "date", "id", "object"]));
    Snake.Logging.log("network", "additionalData: " + Snake.Logging.stringify(
        _(data).omit(["privateData", "date", "id", "object"])));

    // Do not allow messages containing generated fields (i.e. starting
    // with an underscore).
    findUnderscore = function (value, key) {
      return key[0] === "_" ||
          (_.isObject(value) && !_.isUndefined(_.find(value, findUnderscore)));
    };
    if (!_.isUndefined(_.find(data, findUnderscore))) {
      throw new Error("Message contains a field starting with \"_\"");
    };

    return modelClass.handleReceivedRawData(data).then(function () {
      return modelClass.getEncryptionKey(data);
    }).then(function(key) {
      ciphertext = Snake.IO.deserializeBinary(privateData.data);

      if (_(key).isUndefined()) {
        // Couldn't find a key
        if (data.isPublic === true) {
          // OK, it's public
          return ciphertext;
        }
        throw new Error("Decryption key not available");
      }

      iv = Snake.IO.deserializeBinary(privateData.iv);

      return Snake.Crypto.Common.symmAuthDecryption(key, ciphertext, iv,
          additionalData).catch(function (e) {
            var additionalInfo = "Key: " + Snake.Logging.stringify(key) + "\n" +
              "Ciphertext: " + Snake.Logging.stringify(ciphertext) + "\n" +
              "IV: " + Snake.Logging.stringify(iv) + "\n" +
              "Additional: " + Snake.Logging.stringify(additionalData);
            throw new Error(e.message  + "\n" + additionalInfo);
          });
    }).then(function(cleartext) {

      cleartext = msgpack.unpack(cleartext);

      // Expand data with RJSON-unpacked private data
      _.extend(data, RJSON.unpack(cleartext.data));

      // TODO: if (model.hasToVerifySignature)
      if (_(cleartext).has("signature")) {
        dataToVerify = msgpack.pack(data);

        // Get the public key and convert it to JWK
        return modelClass.getSigningPublicKey(data).then(function(publicKey) {
          // Check if we can obtain the reuqired public key
          if (_(publicKey).isUndefined()) {
            throw new Error("No public key available to verify signature");
          }

          // Verify the signature
          return Snake.Crypto.Common.verify(dataToVerify, cleartext.signature,
              publicKey);
        });
      }
    }).then(function() {
      return data;
    });
  }

  /**
   * Validates a model with a given validator.
   * @param {Object} attributes
   * @param {} modelClass
   * @returns {Array} array with all the validation errors.
   */
  function validateModel(attributes, schema, typeName) {
    var result, errors, message;
    result = tv4.validateMultiple(attributes, schema);
    errors = result.errors;
    if (errors.length !== 0) {
      message = "Invalid object id: " + attributes.id +
          ", type: " + typeName;
      _.each(errors, function(error) {
        message += "\n" + error.dataPath + ": " + error.message;
      });

      message += "\n" + Snake.Logging.stringify(attributes, null, "\t");

      Snake.Logging.error("validation", message);
    }

    return errors;
  }

  /**
   * Namespace containing the implementation of methods used by Backbone.sync
   * using SocketIO.
   * @class Snake.IO
   */
  // TODO: all .reject should give an Error
  Snake.IO = _.extend(Backbone.Events, {
    // TODO: do we need this?
    pending: {
      create: 0,
      read: 0,
      update: 0
    },
    init: function() {
      Backbone.sync = function(method, model, options) {
        return Snake.IO[method](model, options);
      };
    },
    serializeBinary: function(data) {
      return window.base91.encode(data);
    },
    deserializeBinary: function(data) {
      return window.base91.decode(data);
    },
    addPending: function(type) {
      this.pending[type]++;
      this.trigger(type + ":" + String(this.pending[type]));
    },
    removePending: function(type) {
      this.pending[type]--;
      this.trigger(type + ":" + String(this.pending[type]));
    },
    /**
     * Stores a new model on the server.
     * @memberOf Snake.IO#
     * @param {Backbone.Model} model the model to store.
     * @param {Object} options object containing:
     *        * error: a callback in case of error (e.g. validation)
     *        * success: a callback in case of success (but please use the
     *          returned promise)
     *        * hash: a Boolean indicating whether to compute the hash of the
     *                message sent or not
     * @returns {Object} object containing two promises:
     *          * result: a Promise resolved when the message has been sent
     *          * hash: a Promise resolved when the hash of the outgoing message
     *            has been computed
     */
    create: function(model, options) {
      return this.modify("create", model, options);
    },
    /**
     * Updates an existing model on the server.
     * @memberOf Snake.IO#
     * @param {Backbone.Model} model the model to update.
     * @param {Object} options object containing:
     *        * error: a callback in case of error (e.g. validation)
     *        * success: a callback in case of success (but please use the
     *          returned promise)
     *        * hash: a Boolean indicating whether to compute the hash of the
     *                message sent or not
     * @returns {Object} object containing two promises:
     *          * result: a Promise resolved when the message has been sent
     *          * hash: a Promise resolved when the hash of the outgoing message
     *            has been computed
     */
    update: function(model, options) {
      return this.modify("update", model, options);
    },
    modify: function(method, model, options) {
      var result, hashResolve, hashReject, payload;

      // The result contains two promises, one for when the hash is ready
      // and the other for when the result of the operation is available
      result = {
        hash: new Promise(function(resolve, reject) {
          hashResolve = resolve;
          hashReject = reject;
        })
      };

      result.result = Snake.identityPromise().then(function() {
        var dataToSend, findLocalFields, validationErrors;

        // We don't want to modify the original elements
        dataToSend = _(model.attributes).deepClone();

        model.constructor.handlePreSendRawData(dataToSend);

        // Check for remaining local properties
        findLocalFields = function (item) {
          _(item).forEach(function (value, key) {
            if (key[0] === "_") {
              throw new Error("Can't send objects containing local fields");
            } else if (_.isObject(value)) {
              findLocalFields(value);
            }
          });
        };
        findLocalFields(dataToSend);

        // TODO: drop date
        if (method === "create") {
          // Sets id and date with fake values just to pass the validation,
          // these values will be overwritten when the server confirms the
          // successful creation of the model
          dataToSend.id = 0;
          dataToSend.date = new Date().toISOString();
        }

        validationErrors = validateModel(dataToSend, getOutgoingSchema(model),
            model.urlRoot);

        if (validationErrors.length > 0) {
          hashReject(validationErrors);
          throw validationErrors;
        }

        Snake.Logging.log("network", "Using \"" + method +
            " method with parameter " + Snake.Logging.stringify(dataToSend));

        if (options.rawRequest) {
          dataToSend = foldData(dataToSend, getSchema(model),
            _.result(model, "urlRoot"));
          dataToSend.privateData = "";
          return dataToSend;
        } else {
          return signAndCrypt(dataToSend, model);
        }
      }.bind(this)).then(function(encryptedModel) {
        payload = encryptedModel;

        // Send request
        this.addPending(method);
        return Snake.Network[method](payload);
      }.bind(this)).then(function(response) {
        Snake.Perf.TrafficCounter.stat("received", method, response);

        // Handle response
        if ("error" in response) {
          // Tries to perform again the operation after a second
          // TODO temporary
          if (response.code === Snake.Config.Network.TooFast) {
            return Snake.slowPokePromise(1000).then(function() {
              return Snake.IO.modify(method, model, options);
            });
          }

          this.removePending(method);
          hashReject(response);
          throw new Error("Error from the server: " +
              Snake.Logging.stringify(response) + "\" while sending\n" +
              Snake.Logging.stringify(payload, null, "\t"));
        }

        // TODO: drop date
        if (method === "create") {
          model.attributes.date = response.date;
          model.set("id", response.id);
        }

        if (options.hash === true) {
          delete payload.object;
          payload.date = response.date;
          if (method === "create") payload.id = response.id;
          hashResolve(hashModel(payload));
        }

        Snake.Logging.log("network", "Response from the server: " +
            Snake.Logging.stringify(response));
        this.removePending(method);
        options.success({});
      }.bind(this)).catch(Snake.errorHandler.bind("network"));

      return result;
    },
    /**
     * Reads a model from the server (or the local cache).
     * @memberOf Snake.IO#
     * @param {Backbone.Model} model the model to fetch.
     * @param {Object} options object containing:
     *        * error: a callback in case of error (e.g. validation)
     *        * success: a callback in case of success (but please use the
     *          returned promise)
     *        * hash: a Boolean indicating whether to compute the hash of the
     *                message sent or not
     * @returns {Object} object containing two promises:
     *          * result: a Promise resolved when the message has been sent
     *          * hash: a Promise resolved when the hash of the outgoing message
     *            has been computed
     */
    read: function(model, options) {
      var payload, cached, result, hashResolve, hashReject, hash, rawResponse,
          modelClass, parsingPromises;

      result = {
        hash: new Promise(function(resolve, reject) {
          hashResolve = resolve;
          hashReject = reject;
        })
      };

      result.result = Snake.identityPromise().then(function() {
        var hashes, binarySerializer;

        // TODO: validate a schema also for requests
        payload = _.result(model, "url");
        // TODO: and use binary formats to detect which fields must be binary
        //       serialized, as we do in foldData
        binarySerializer = function (obj) {
          _(obj).forEach(function (value, key) {
            if (value instanceof Uint8Array) {
              obj[key] = Snake.IO.serializeBinary(obj[key]);
            } else if (_.isObject(value)) {
              binarySerializer(value);
            }
          });
        };
        binarySerializer(payload);

        Snake.Logging.log("network",
            "Using \"read\" method with parameter " +
            Snake.Logging.stringify(payload));

        cached = this.Cache.get(payload, options.hash);
        if (!_.isUndefined(cached)) {
          Snake.Logging.log("network", "Cache hit!");

          if (options.hash === true) {
            hashResolve(cached.hash);
          }

          options.success && options.success(cached.result);
          return cached.result;
        }

        // It's not cached! Send the request to the server
        this.addPending("read");
        return Snake.Network.read(payload).then(function(response) {
          Snake.Perf.TrafficCounter.stat("received", "read", response);

          var parse, saveHash;

          if ("error" in response) {
            this.removePending("read");
            if (options.hash === true) hashReject(response);
            throw new Error("Error from the server: \"" +
                Snake.Logging.stringify(response) + "\" while sending\n" +
                Snake.Logging.stringify(payload, null, "\t"));
          }

          // Save for hashing
          rawResponse = _(response).deepClone();

          if (options.rawResponse) {
            return rawResponse;
          }

          parse = function(responseItem) {
            var schemaName;

            schemaName = Snake.Schema.urlRootToSchemaName(payload.object);

            // Find the class (constructor) responsible for the current request
            modelClass = _(Snake.Model).find(function(classModel) {
              return _.has(classModel, "isMyDuty")
                  && classModel.isMyDuty(schemaName, responseItem);
            });

            if (_.isUndefined(modelClass)) {
              throw new Error("Couldn't find an appropriate model for the " +
                  "object type \"" + schemaName + "\"");
            }

            // Compute the hash and notify those who need it
            if (options.hash === true) {
              hash = hashModel(responseItem);

              // TODO: this is ugly
              saveHash(hash);
            }

            return decryptAndVerify(responseItem, modelClass)
                .catch(function (e) {
              Snake.Logging.warn("network", "Error while decrypting the "
                + "following:\n" + Snake.Logging.stringify(responseItem)
                + "\nIn response to the following request:\n"
                + Snake.Logging.stringify(payload) + "\n" + e.message);
              throw e;
            }).then(function(
                attributes) {
              var validationErrors;

              Snake.Logging.log("network", "Response from the server: " +
                  Snake.Logging.stringify(attributes));

              validationErrors = validateModel(attributes, modelClass);
              if (validationErrors.length > 0) {
                this.removePending("read");
                if (options.hash === true) hashReject(attributes);
                throw validationErrors;
              }

              // Cache flat & decrypted data (+ hash)
              this.Cache.add(payload, attributes, hash);

              return attributes;
            }.bind(this));
          }.bind(this);

          // TODO: do we ever have non-array responses from the server?
          if (_(rawResponse).isArray()) {

            if (rawResponse.length === 0) {
              hashes = Snake.identityPromise([]);
              return rawResponse;
            }
            saveHash = function(newHash) {
              if (_(hashes).isUndefined()) hashes = [];
              hashes.push(newHash);
            };

            // Make failure of parsing a single message non-critical
            parsingPromises = _(rawResponse).map(parse);
            parsingPromises = _(parsingPromises).map(function (promise) {
              return promise.catch(function (error) {
                console.warn(error);
              });
            });

            return Promise.every.apply(this, parsingPromises).then(
                function(parsed) {
                  var i;

                  // Remove undefined entries and corresponding hashes
                  for (i = 0; i < parsed.length; i++) {
                    if (_.isUndefined(parsed[i])) {
                      parsed.splice(i, 1);
                      if (!_.isUndefined(hashes))
                        hashes.splice(i, 1);
                      i--;
                    }
                  }

                  hashes = Promise.every.apply(this, hashes);
                  return parsed;
                });
          } else {
            saveHash = function(hash) {
              hashes = hash;
            };
            return parse(rawResponse);
          }
        }.bind(this)).then(function(parsed) {
          this.removePending("read");

          if (options.hash === true && !_(hashes).isUndefined()) {
            hashes.then(function(res) {
              hashResolve(res);
            });
          }

          options.success && options.success(parsed);
          return parsed;
        }.bind(this));
      }.bind(this)).catch(Snake.errorHandler.bind("network"));

      return result;
    },
    /**
     * Deletes an existing model.
     * @memberOf Snake.IO#
     * @param {Backbone.Model} model the model to delete.
     * @returns {Object} object containing a Promise resolved when the message
     *                   has been deleted
     */
    delete: function(model) {
      // TODO implement me
      return Snake.identityPromise();
    }
  });

  /**
   * Class for transparently caching server's responses
   * @class Snake.IO.Cache
   */
  Snake.IO.Cache = {
    stats: {
      hits: 0,
      misses: 0
    },
    cache: {},
    hashCache: {},
    /**
     * Adds url-response(+hash) pairs to the cache
     * @memberOf Snake.IO.Cache#
     * @param {Object} url the key, i.e. the request payload.
     * @param {Object} response the associated response.
     * @param {Object} hash hash of the message to cache.
     */
    add: function(url, item, hash) {
      if (!this.isCachable(url)) return;

      // Ensure initialization
      this.cache[url.object] = this.cache[url.object] || {};
      this.hashCache[url.object] = this.hashCache[url.object] || {};

      // Cache the element of the response
      if (!_.contains(url.ids, item.id)) {
        Snake.Logging.warn("network", "Got a non requested id.");
        return;
      }

      this.cache[url.object][item.id] = item;
      if (!_(hash).isUndefined()) this.hashCache[url.object][item.id] = hash;

    },
    /**
     * Looks in the cache for all the supplied ids, if all of them are
     * available return the corresponding responses in an array, otherwise
     * return undefined.
     * @memberOf Snake.IO.Cache#
     * @param {Object} url the payload of the request.
     * @param {Boolean} hash whether the hash of the message is needed too.
     * @return {Object} array of the objects associated with the requested
     *         ids or undefined in case the match is not perfect. The objects
     *         will have the following fields
     *         * result: the actual cached response
     *         * hash (optional): the hash of the message associated with the
     *           response
     */
    get: function(url, hash) {
      var result = {};

      hash = _(hash).isUndefined() ? false : hash;

      if (!this.isCachable(url)) return;

      if (!_(this.cache).has(url.object)) {
        this.stats.misses++;
        return;
      }

      // Take all the corresponding response from the cache
      // Note: undefined here are fine, see check afterwards
      result.result = url.ids.map(function(id) {
        // We give back a clone, we don't want objects in cache to
        // be modified. Clone should be enough since the cache holds
        // folded data
        return _(this.cache[url.object][id]).clone();
      }.bind(this));

      // Even if a single one is missing, return nothing
      // TODO: implement merge of cached and fetch'd results
      if (result.result.indexOf(undefined) !== -1) {
        this.stats.misses++;
        return;
      }

      if (hash) {
        result.hash = url.ids.map(function(id) {
          return this.hashCache[url.object][id];
        });
        if (result.hash.indexOf(undefined) !== -1) {
          this.stats.misses++;
          return;
        }
      }

      this.stats.hits++;
      return result;
    },
    /**
     * Checks if it's a cachable query, i.e. only contains the ids and
     * objects fields
     * @memberOf Snake.IO.Cache#
     * @param {Object} url the request payload to check.
     * @returns {Boolean} true if it should be managed by the cache.
     */
    isCachable: function(url) {
      return _.difference(_(url).keys(), ["ids", "object"]).length === 0;
    },
    clearStats: function() {
      this.stats = {
        hits: 0,
        misses: 0
      };
    },
    getStats: function() {
      var hitRate = this.stats.hits / (this.stats.hits + this.stats.misses);
      var missRate = 1 - hitRate;
      return {
        hitRate: hitRate,
        missRate: missRate
      };
    },
    /**
     * Cleans the cache.
     * @memberOf Snake.IO.Cache#
     * @param {String} [type] removes only objects of the specified
     *        type from the cache.
     */
    flush: function(type) {
      if (_.isUndefined(type)) {
        this.cache = {};
        this.hashCache = {};
      } else {
        delete this.cache[type];
        delete this.hashCache[type];
      }
    }
  };
}());
