(function() {
  "use strict";

  Snake.Modal = _.extend(Snake.Modal || {}, {
    init: function() {
      this.element = $("#modal-dialog");
      this.title = this.element.find(".modal-title");
      this.body = this.element.find(".modal-body");
      this.confirmButton = this.element.find(".btn-primary");
      this.cancelButton = this.element.find(".btn-default");

      // Cancel
      this.element.find(".btn-default, .close").on("click", function() {
        this.element.one('hidden.bs.modal', function() {
          this.resolve();
        }.bind(this));
        this.element.modal("hide");
      }.bind(this));

      // Confirm
      this.confirmButton.on("click", function() {
        var result;

        result = this.body.serializeArray().reduce(function(res, val) {
          res[val.name] = val.value;
          return res;
        }.bind(this), {});

        this.element.one('hidden.bs.modal', function() {
          this.resolve(result);
        }.bind(this));

        this.element.modal("hide");

      }.bind(this));

      this.lastComplete = Snake.identityPromise();
    },
    confirm: function(title, message) {
      var body;

      body = Snake.instanceTemplate("confirm-modal", {message: message});
      return this.enqueue(title, body, "Yes", "No")
          .then(function(result) {
        return !_(result).isUndefined() && result.confirm === "yes";
      });
    },
    enqueue: function(title, body, confirmText, cancelText) {
      var resultResolve, resultReject, newPromise;

      // Create a new promise
      newPromise = new Promise(function(resolve, reject) {
        resultResolve = resolve;
        resultReject = reject;
      }.bind(this));

      // When the last promise completes, show the new dialog
      this.lastComplete.then(function() {
        this.title.text(title);
        this.body.empty();
        this.body.append(body);
        this.confirmButton.text(confirmText || "OK");
        this.cancelButton.text(cancelText || "Cancel");

        // And update the current resolver
        this.resolve = resultResolve;
        this.reject = resultReject;
        this.element.modal("show");

        // Now wait for pressure of a button
      }.bind(this));

      // Substitute the last promise with the new one
      this.lastComplete = newPromise;

      return newPromise;
    }
  });

}());
