(function() {
  "use strict";

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for a post
     * @class Snake.Model.Post
     */
    Post: Snake.Model.ModelBase.extend({
      childModels: ["author", "recipient"],
      initialize: function(attributes, options) {
        this.options = _.defaults(options || {}, {
          fetchAuthor: false,
          fetchRecipient: false,
          author: {},
          recipient: {}
        });

        this.nowAndOn(["_recipientType"], function() {
          if (this.has("_recipientType") &&
              this.get("_recipientType") !== "group") {
            throw new Error("Using post model but _recipientType is \"" +
              this.get("_recipientType") + "\"");
          }
        }.bind(this));

        // Fetch the group of friends associated to this profile
        if (this.options.fetchAuthor) {
          this.author = new Snake.Model.User({}, this.options.author);
          this.monitorChange("author", "authorId");
        }

        if (this.options.fetchRecipient) {
          this.recipient = new Snake.Model.Group({}, this.options.recipient);

          this.monitorChange("recipient", "_recipientId");
        }
      }
    }, {
      urlRoot: "message",
      // TODO: factorize with message.js
      handlePreSendRawData: function (data) {
        var match;

        delete data._fakeId;

        // If a message has an address manually specified it shouldn't have
        // local fields to deduce it
        if (_.has(data, "address")) {
          if (_.has(data, "_senderId")) delete data._senderId;
          if (_.has(data, "_recipientId")) delete data._recipientId;
          if (_.has(data, "_recipientType")) delete data._recipientType;
        } else if (_.has(data, "_recipientType") ||
                   _.has(data, "_recipientId")) {
          // We have to obtain the correct address given the specified local
          // fields
          if (data._recipientType === "group") {
            // User to group
            match = Snake.sessionUser.getGroup(data._recipientId);
            data.address = match.address;
          } else {
            throw new Error("Unknown _recipientType: " + data._recipientType);
          }

          delete data._recipientType;
          delete data._recipientId;
        }
      },
      /// A message is a post only if the address is the address of known group
      isMyDuty: function(urlRoot, partialData) {
        var match, address;

        if (!Snake.Model.ModelBase.isMyDuty.call(this, urlRoot))
          return false;

        if (!_.has(partialData, "address"))
          return false;

        address = base91.decode(partialData.address);

        match = _(Snake.sessionUser.get("groups"))
            .findByField(address, "address");

        return !_.isUndefined(match);
      },
      handleReceivedRawData: function(data) {
        var match, address, sessionUser;

        // TODO: move Snake.Logging.stringify in utils.js?
        // TODO: what's the whole point of _fakeId?
        data._fakeId = data.id.toString() + "-" +
          Snake.Logging.stringify(data.address);

        if (!_(data).has("address"))
          throw new Error("The post doesn't have a destination address");

        address = data.address;

        // User to group, group is the recipient
        match = _(Snake.sessionUser.get("groups"))
            .findByField(address, "address");

        if (!_.isUndefined(match)) {
          data._recipientType = "group";
          data._recipientId = match.id;
          return Snake.identityPromise();
        } else {
          throw new Error("Received a message for an unknown address");
        }
      },
      getEncryptionKey: function(data) {
        var groups = _(Snake.sessionUser.get("groups"));
        var group = groups.findByField(data._recipientId);

        if(_(data.id).isUndefined()
           || data.id === 0
           || data.id > group.lastValidMessageId) {
          // the message can be decrypted with the current group key
          var key = group.keys.rootKey.key;
          return Snake.identityPromise(key);
        } else {
          // the message cannot be decrypted with the current group key, fetch
          // the group descriptor in order to get the proper key to decrypt it
          var groupModel = new Snake.Model.Group({id: group.id});
          return groupModel.fetch().result.then(function(res) {
            var oldKeys = res.oldKeys;
            for (var i = 0; i < oldKeys.length; i++) {
              if(data.id <= oldKeys[i].lastValidMessageId)
                return oldKeys[i].key;
            }
          });
        }
      },
      getSigningPublicKey: function(data) {
        return Snake.sessionUser.getUserSigningKey(data.authorId);
      }
    })
  });

}());
