(function() {
  "use strict";

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for the user profile
     * @class Snake.Model.Profile
     */
    Profile: Snake.Model.ModelBase.extend({
      childModels: ["friendsGroup", "owner"],
      defaultValues: {
        name: "",
        surname: "",
        email: "",
        birthday: "1970-01-01",
        groups: [],
        priority: 0
      },
      initialize: function(attributes, options) {
        this.options = _.defaults(options || {}, {
          setDefaultValues: false,
          fetchFriendsGroup: false,
          fetchGroups: false,
          fetchOwner: false,
          groups: {},
          friendsGroup: {},
          owner: {}
        });

        // Copies defaults values into the model
        if (this.options.setDefaultValues) {
          this.loadDefaultValues();
        }

        if (this.options.fetchFriendsGroup) {
          this.friendsGroup = new Snake.Model.Group({},
              this.options.friendsGroup);
        }

        if (this.options.fetchOwner) {
          this.owner = new Snake.Model.User({}, this.options.owner);
          this.monitorChange("owner", "userId");
        }

        if (this.options.fetchGroups ||
            this.options.fetchFriendsGroup) {
          this.groups = new Snake.Model.Collection([], {
            model: Snake.Model.Group,
            url: function() {
              if (!this.has("groups") || this.get("groups").length === 0) return;
              return {
                object: "group",
                ids: _(this.get("groups")).pluck("id")
              };
            }.bind(this)
          });

          this.groups.on("add remove", function() {
            var friendsGroups;

            // TODO: create a mixin for this?
            // Choose the friend's group
            friendsGroups = _(this.groups.models).filter(function(group) {
              return (group.get("friendsOf") === this.get("userId"));
            }.bind(this));

            // TODO: what if there's more than one?
            if (friendsGroups.length > 0 &&
                this.friendsGroup.id !== friendsGroups[0].id) {

              this.friendsGroup.setId(friendsGroups[0].id);
              this.friendsGroup.fetch(this.options.friendsGroup);
            }
          }.bind(this), this);

          this.monitorChange("groups");
        }
      }
    }, {
      urlRoot: "profile",
      getEncryptionKey: function(data) {
        var result, id;
        id = data.id;
        _(Snake.sessionUser.get("profiles")).each(function(profile) {
          if (profile.id === id) result = profile.key;
        });

        if (!_(result).isUndefined()) return Snake.identityPromise(result);

        _(Snake.sessionUser.get("friends")).each(function(friend) {
          var matching;
          matching = _(friend.profile).filter(function(profile) {
            return profile.id === id;
          });

          if (matching.length > 0) {
            if (!_(result).isUndefined() || matching.length > 1) {
              throw new Error("Too many matches");
            }

            result = matching[0];
          }
        });

        return Snake.identityPromise(result);
      },
      getSigningPublicKey: function(data) {
        return Snake.sessionUser.getUserSigningKey(data.userId);
      }
    })
  });
}());
