(function() {
  "use strict";

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for the user
     * @class Snake.Model.User
     */
    User: Snake.Model.ModelBase.extend({
      childModels: ["actualProfile", "messages"],
      defaultValues: {
        publicProfile: 0,
        groups: [],
        friends: [],
        profiles: []
      },
      initialize: function(attributes, options) {
        // By default fetch the profile
        this.options = _.defaults(options || {}, {
          setDefaultValues: false,
          fetchActualProfile: false,
          fetchMessages: 0,
          actualProfile: {},
          messages: {}
        });

        // Copies defaults values into the model
        if (this.options.setDefaultValues) {
          this.loadDefaultValues();
        }

        // Initialize an empty profile
        if (this.options.fetchActualProfile) {
          this.actualProfile = new Snake.Model.Profile({userId: this.id},
          this.options.actualProfile);

          // TODO: can we make monitorChange general enough to handle this?

          // We default to use the public profile
          this.on("change:publicProfile", this.newActualProfile, this);

          // TODO: we can't listen to friends[id].profiles, is this OK?
          Snake.sessionUser.on("change:friends", this.newActualProfile, this);
          this.newActualProfile();
        }

        if (this.options.fetchMessages > 0) {
          this.messages = new Snake.Model.Collection([], {
            model: Snake.Model.Message,
            comparator: Snake.dateComparator(),
            url: function () {
              var friend = Snake.sessionUser.getFriend(this.id);
              return {
                object: "last-messages",
                addresses: [friend.myAddress, friend.otherAddress],
                num: this.options.fetchMessages
              };
            }.bind(this)
          });
          this.fetchNestedModel("messages");

          // TODO: create an event to notify of new messages
        }
      },
      newActualProfile: function() {
        var friend, profileId;

        // Check if we are friends
        if (Snake.sessionUser.has("friends")) {
          friend = _(Snake.sessionUser.get("friends")).findByField(this.id);

          if (!_.isUndefined(friend) && _(friend).has("profiles")) {
            // TODO: allow to choose in case multiple profiles are avilable
            // OK, we have at least a profile
            profileId = friend.profiles[0].profileId;
          }
        }

        // We don't have profile reference in our friend list, use the
        // public profile
        if (_(profileId).isUndefined() && this.has("publicProfile")) {
          profileId = this.get("publicProfile");
        }

        // If we were able to find a profile ID, use it
        if (!_(profileId).isUndefined()) {
          this.actualProfile.setId(profileId);
          this.fetchNestedModel("actualProfile");
        }
      },
      /**
       * Returns the data that need to be authenticated to verify the identity
       * of the user.
       * @memberOf Snake.Model.User#
       * @returns {ArrayBufferView} the requested data.
       */
      getDataToAuthenticate: function() {
        var id;
        // TODO: include all public fields?
        id = this.id.toString(16);
        id = ((id.length % 2 !== 0) ? "0" : "") + id;
        id = util.hex2abv(id);

        return Snake.concatBuffers(this.get("staticPublicKey"), id);
      }
    }, {
      urlRoot: "user",
      isMyDuty: function(urlRoot, partialData) {
        return Snake.Model.ModelBase.isMyDuty.call(this, urlRoot);
      },
      getSigningPublicKey: function(data) {
        return Snake.sessionUser.getUserSigningKey(data.id);
      }
    })
  });
}());
