(function() {
  Snake.Perf = {};

  Snake.Perf.CryptoCounter = {
    stats: {},
    stat: function(type, count, start, end) {
      if (!this.stats.hasOwnProperty(type)) {
        this.stats[type] = {count: [], startTime: [], endTime: []};
      }
      this.stats[type].count.push(count);
      this.stats[type].startTime.push(start);
      this.stats[type].endTime.push(end);
    },
    getStats: function() {
      var summary = {};
      for (var stat in this.stats) {
        if (this.stats.hasOwnProperty(stat)) {
          var currentStat = this.stats[stat];

          var data = currentStat.count;

          var times = [];
          for (var i = 0; i < currentStat.startTime.length; i++) {
            times[i] = currentStat.endTime[i] - currentStat.startTime[i];
          }

          var totalBytes = data.reduce(function(val, acc) {
            return val + acc;
          }, 0);

          var totalTime = times.reduce(function(val,acc) {
            return val+acc;
          }, 0);

          summary[stat] = {
            count: data.length,
            totalBytes: totalBytes,
            data: data,
            totalTime: totalTime,
            time: times
          };
        }
      }
      return summary;
    },
    getRealTime: function() {
      var times = [];

      for (var stat in this.stats) {
        if (this.stats.hasOwnProperty(stat)) {
          var currentStat = this.stats[stat];
          for (var i = 0; i < currentStat.startTime.length; i++) {
            times.push({
              startTime: currentStat.startTime[i],
              endTime: currentStat.endTime[i]
            });
          }
        }
      }

      times = times.sort(function(a, b) {
        return a.startTime - b.startTime;
      });

      var i = 0;
      do {
        if (times[i].endTime > times[i+1].startTime) {
          times[i].endTime = Math.max(times[i].endTime, times[i+1].endTime);
          times.splice(i+1, 1);
        } else {
          i++;
        }
      } while (i < times.length - 1);

      var realTime = times.reduce(function(acc, val) {
        return acc + val.endTime - val.startTime;
      }, 0);

      return realTime;
    },
    flush: function() {
      this.stats = {};
    }
  };

  Snake.Perf.TrafficCounter = {
    sentStats: {},
    receivedStats: {},
    stat: function(direction, type, data) {
      var stats;
      if (direction === "sent") {
        stats = this.sentStats;
      } else if (direction === "received") {
        stats = this.receivedStats;
      }

      if (!stats.hasOwnProperty(type)) {
        stats[type] = [];
      }

      var size = JSON.stringify(data).length;
      stats[type].push(size);
    },
    getStats: function() {
      return {
        sent: this.sentStats,
        received: this.receivedStats
      };
    },
    flush: function() {
      this.sentStats = {};
      this.receivedStats = {};
    }
  };

}());
