(function() {
  "use strict";

  /**
   * Routing class, decides which page to show from the URL fragment
   * @class Snake.Router
   */
  Snake.Router = Backbone.Router.extend({
    initialize: function(options) {
      this.model = options.model;
      this.view = options.view;
    },
    routes: {
      "profile": "profile",
      "profile/:id": "profile",
      "manageprofiles": "manageProfiles",
      "managegroups": "manageGroups",
      "managefriends": "manageFriends",
      "group/:id": "group",
      "groupadmin/:id": "groupAdmin",
      "messages/:id": "messages",
      "stream": "stream"
    },
    createPage: function(modelId, modelOptions, backModel, backView, pagePath) {
      // If the page has not been loaded yet create model and view
      if (!(pagePath in this.model.pages)) {
        // Select what data we want to show
        var model = new backModel({
          id: modelId
        }, modelOptions);

        this.model.addPage(pagePath, model);

        // Create views
        this.view.addPage(pagePath, new backView({
          model: model
        }));
      }

      this.model.changePage(pagePath);
    },
    stream: function() {
      this.createPage(Snake.sessionUser.id, {
        fetchGroups: true,
        groups: {
          fetchFriendsOf: true,
          fetchMessages: 10,
          messages: {
            fetchAuthor: true,
            fetchRecipient: true,
            recipient: {
              fetchFriendsOf: true,
              friendsOf: {
                fetchActualProfile: true
              }
            },
            author: {
              fetchActualProfile: true
            }
          }
        }
      }, Snake.Model.SessionUser, Snake.View.Stream, "stream");
    },
    messages: function(id) {
      id = new Number(id).valueOf();
      var pagePath = "messages/" + id.toString();

      this.createPage(id, {
        fetchActualProfile: true,
        fetchMessages: 10,
        messages: {
          fetchSender: true,
          sender: {
            fetchActualProfile: true
          }
        }
      }, Snake.Model.User, Snake.View.Messages, pagePath);

      // Refresh messages
      Snake.sessionUser.processNewMessages();
    },
    profile: function(id) {
      id = id || Snake.sessionUser.id;
      id = new Number(id).valueOf();
      var pagePath = "profile/" + id.toString();

      this.createPage(id, {
        fetchOwner: true,
        fetchFriendsGroup: true,
        friendsGroup: {
          fetchMembers: true,
          fetchAdministrators: true,
          fetchMessages: 10,
          members: {fetchActualProfile: true},
          administrators: {fetchActualProfile: true},
          messages: {
            fetchAuthor: true,
            author: {
              fetchActualProfile: true
            }
          }
        }
      }, Snake.Model.Profile, Snake.View.Profile, pagePath);
    },
    manageProfiles: function() {
      this.createPage(Snake.sessionUser.id, {
        fetchProfiles: true
      }, Snake.Model.SessionUser, Snake.View.ManageProfiles, "manageprofiles");
    },
    manageGroups: function() {
      this.createPage(Snake.sessionUser.id, {
        fetchGroups: true,
        groups: {
          fetchMembers: true,
          fetchAdministrators: true
        }
      }, Snake.Model.SessionUser, Snake.View.ManageGroups, "managegroups");
    },
    manageFriends: function() {
      this.createPage(Snake.sessionUser.id, {
        fetchFriends: true,
        friends: {
          fetchActualProfile: true
        }
      }, Snake.Model.SessionUser, Snake.View.ManageFriends, "managefriends");
    },
    group: function(id) {
      id = new Number(id).valueOf();
      var pagePath = "group/" + id.toString();

      this.createPage(id, {
        fetchMembers: true,
        fetchAdministrators: true,
        fetchMessages: 10,
        members: {fetchActualProfile: true},
        administrators: {fetchActualProfile: true},
        messages: {
          fetchAuthor: true,
          author: {
            fetchActualProfile: true
          }
        }
      }, Snake.Model.Group, Snake.View.Group, pagePath);
    },
    groupAdmin: function(id) {
      id = new Number(id).valueOf();
      var pagePath = "groupadmin/" + id.toString();

      this.createPage(id, {
        fetchMembers: true,
        fetchAdministrators: true,
        fetchAdminData: true,
        members: {fetchActualProfile: true},
        administrators: {fetchActualProfile: true}
      }, Snake.Model.Group, Snake.View.GroupAdmin, pagePath);
    }

  });

}());
