(function() {
  "use strict";

  var directives = {
    success: {
      class: function(params) {
        if (this.formValid) {
          return params.value;
        } else {
          return params.value + " hide";
        }
      }
    },
    error: {
      class: function(params) {
        if (this.formValid === false) {
          return params.value;
        } else {
          return params.value + " hide";
        }
      }
    },
    fields: {
      field: {
        "class": function(params) {
          if (this.invalid) {
            return params.value + " has-error";
          } else {
            return params.value;
          }
        }
      },
      label: {
        "for": function(params) {
          // TODO add an id for each form field?
        },
        "text": function() {
          return this.title;
        }
      },
      input: {
        "id": function() {
          // TODO add an id for each form field?
        },
        "type": function() {
          return this.input;
        },
        "name": function() {
          return this.name;
        },
        "placeholder": function() {
          return this.description;
        },
        "value": function() {
          return this.value || "";
        }
      },
      message: {
        "text": function() {
          if (this.invalid) {
            return this.error;
          }
        },
        "class": function(params) {
          if (this.invalid) {
            return params.value;
          } else {
            return params.value + " hide";
          }
        }
      }
    }
  };

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for a Form.
     * @class Snake.View.Form
     */
    Form: Snake.View.ViewBase.extend({
      events: {
        "submit": "submit",
        "click .close": "close"
      },
      initialize: function(options) {
        options = _.defaults(options || {}, {
          legend: "Form",
          button: "Submit",
          beforeSave: Snake.identityPromise,
          schema: {}
        });

        this.legend = options.legend;
        this.button = options.button;
        this.beforeSave = options.beforeSave;
        var schema = options.schema;

        // Creates an array with the form fields based on the schema
        this.fields = [];
        _.each(schema.formFields, function(field) {
          this.fields.push(_.extend({name: field}, schema.properties[field]));
        }.bind(this));
      },
      changeModel: function(model) {
        // Updated the model
        this.model = model;

        // Sets all the fields as valid
        _.each(this.fields, function(value) {
          value.invalid = false;
        });
        // Delete the infomation of form validity, this way neither the
        // success message nor the error message are rendered
        delete this.formValid;

        this.render();
      },
      render: function() {
        var newForm;

        // Copies values of the model into fields if it is available
        if(this.model) {
          this.fields = _.map(this.fields, function(field) {
            field.value = this.model.get(field.name);
            return field;
          }, this);
        }

        var formData = {
          legend: this.legend,
          fields: this.fields,
          button: this.button,
          formValid: this.formValid
        };

        newForm = Snake.instanceTemplate("form", formData, this.el.id,
            directives);

        this.$el.replaceWith(newForm);
        this.setElement(newForm[0]);

        return this;
      },
      submit: function(event) {
        event.preventDefault();

        // Puts submitted data into a dictionary
        var form = this.$el.find("form");
        var formData = _(form.serializeArray()).reduce(function(acc, field) {
          acc[field.name] = field.value;
          return acc;
        }, {});

        _.each(formData, function(value, key) {
          this.model.set(key, value);
        }, this);

        return this.beforeSave(this.model, formData).then(function() {
          return this.model.save({}).result;
        }.bind(this)).then(function() {
            // All the fields and the form are valid
            _.each(this.fields, function(value) {
              value.invalid = false;
            });
            this.formValid = true;

            // Renders the form to show the success message
            this.render();

            this.trigger("form:success");
          }.bind(this),
          function(response) {
            // Sets the flag invalid, with the relative error message,
            // to all the fields that didn't pass the validation. Also
            // sets the form as invalid.
            _.each(this.fields, function(value) {
              var error = _.find(response, {dataPath: "/" + value.name});
              if (error) {
                value.invalid = true;
                value.error = error.message;
              }
            });
            this.formValid = false;

            // Renders the form to show invalid fields and the error message
            this.render();

            this.trigger("form:error");
          }.bind(this)
        );
      },
      close: function(event) {
        this.trigger("form:close");
      }
    })
  });
}());
