(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for a Group.
     * @class Snake.View.Group
     */
    Group: Snake.View.ViewBase.extend({
      nestedViews: ["posts", "administrators", "members", "editor"],
      initialize: function(options) {
        this.administrators = new Snake.View.UpdatingCollection({
          collection: this.model.administrators,
          childViewConstructor: Snake.View.CompactUser
        });
        this.members = new Snake.View.UpdatingCollection({
          collection: this.model.members,
          childViewConstructor: Snake.View.CompactUser
        });
        this.posts = new Snake.View.UpdatingCollection({
          collection: this.model.messages,
          childViewConstructor: Snake.View.Post,
          renderFilter: function(post) {
            return post.get("type") === "post";
          }
        });

        this.editor = new Snake.View.PostEditor({
          collection: this.model.messages,
          recipient: this.model.id,
          post: {
            fetchAuthor: true,
            author: {
              fetchActualProfile: true
            }
          }
        });

        this.listenTo(this.model, "change", this.render);
      },
      render: function() {
        // Create a new instance the template
        var newGroup;

        newGroup = Snake.instanceTemplate("group",
            this.model.attributeTree(), this.el.id);

        this.renderAllNestedViews(newGroup);

        this.$el.replaceWith(newGroup);
        this.setElement(newGroup[0]);

        this.trigger("title", this.title());

        return this;
      },
      title: function() {
        return this.model.get("name") + " group";
      }
    })
  });

}());
