(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View used to manage all the groups.
     * @class Snake.View.GroupAdmin
     */
    GroupAdmin: Snake.View.ViewBase.extend({
      events: {
        "submit": "addMember"
      },
      nestedViews: ["pendingRequests", "administrators", "members"],
      initialize: function(options) {
        this.administrators = new Snake.View.UpdatingCollection({
          collection: this.model.administrators,
          childViewConstructor: Snake.View.ListUsers,
          childViewOptions: {template: "group-administrators"}
        });

        this.members = new Snake.View.UpdatingCollection({
          collection: this.model.members,
          childViewConstructor: Snake.View.ListUsers,
          childViewOptions: {template: "group-members"}
        });

        var members = this.model.members;
        members.bind("item:remove", this.removeMember, this);
        members.bind("item:promoteAdministrator", this.promoteAdministrator, this);
      },
      addMember: function(e) {
        e.preventDefault();

        // TODO something better
        var username = this.$el.find("input")[0].value;

        Snake.IO.read({
          url: {object: "username", username: username}
        }, {rawResponse: true, success: function() {}}).result
            .then(function(response) {
          var newUserId = response[0].id;
          var newUserKey = response[0].key;

          this.model.insertMember(newUserId, newUserKey).then(function() {
            this.showAlertSuccess("The user " + username +
              " has been invited to the group");
          }.bind(this)).catch(function(error) {
            Snake.Logging.error("messaging", error);
            this.showAlertError("Something went wrong...");
          });
        }.bind(this), function(error) {
          Snake.Logging.error("messaging", error);
          this.showAlertError("The user " + username + " does not exists");
        }.bind(this));
      },
      removeMember: function(args) {
        var id = args.id;

        this.model.removeMember(id).then(function() {
            this.showAlertSuccess("The user has been removed from the group");
          }.bind(this)).catch(function(error) {
            Snake.Logging.error("messaging", error);
            this.showAlertError("Something went wrong...");
          }.bind(this));
      },
      promoteAdministrator: function(args) {
        var id = args.id;

        this.model.promoteMemberToAdministrator(id).then(function() {
          this.showAlertSuccess("The user has been promoted to administrator");
        }.bind(this)).catch(function(error) {
          Snake.Logging.error("messaging", error);
          this.showAlertError("Something went wrong...");
        }.bind(this));
      },
      render: function() {
        var newGroupAdmin;

        newGroupAdmin = Snake.instanceTemplate("group-admin",
            this.model.attributeTree(), this.el.id);

        this.renderAllNestedViews(newGroupAdmin);

        this.$el.replaceWith(newGroupAdmin);
        this.setElement(newGroupAdmin[0]);

        this.alertSuccess = this.$el.find(".alert-success");
        this.alertError = this.$el.find(".alert-error");

        this.trigger("title", this.title());

        return this;
      },
      title: function() {
        return this.model.get("name") + " administration page";
      }
    })
  });
}());
