(function() {
  "use strict";

  var Modal = Snake.Modal;

  Snake.View = _.extend(Snake.View || {}, {
    init: function() {
      // Decoupling happens here!
      // To avoid calling View methods from the Crypto module we define here the
      // coupled parts (e.g. modal confirmations) that will be called from there.
      // This also makes testing easier.

      Snake.Crypto.ManualAuthenticator.askUser = function(name, pubKey) {
        return Modal.confirm("Is " + pubKey + " the pubkey of " + name + "?");
      };

      Snake.Model.SessionUser.chooseProfiles = function (profileData) {
        var body;

        // Build the body of the modal dialog
        body = Snake.instanceTemplate("profile-choice-modal",
            profileData, undefined, {
              allow: {
                name: function() {
                  return "allow" + this.id;
                }
              }
            });

        // Show the modal dialog
        return Modal.enqueue("Choose which profiles to show", body)
            .then(function(result) {
              // Collect the chosen IDs
              return _(_(result).map(function(value,
                  key) {
                if (value === "on" && key.indexOf("allow") === 0) {
                  return Number(key.substring(5)).valueOf();
                }
              })).filterUndefined();
            });
      }

      /**
       * Asks the user a question and the relative answer.
       *
       * @param {String} username the username of the friend which will receive
       *                          the question
       *
       * @returns {Object} `undefined` if the user didn't provide any question,
       *                   otherwise an object with the following fields:
       *
       *                       {
       *                         "myQuestion": "...",
       *                         "myAnswer": "..."
       *                       }
       *
       */
      Snake.Crypto.SMP.askQuestion = function (username) {
        return Modal.enqueue("Friend authentication",
                             Snake.instanceTemplate("friend-question-modal",
                                                    {"friend-name": username}));
      };

      /**
       * Asks the user the answer to a question and, if the user wants a new
       * question and the relative answer.
       *
       * @param {String} username the username of the friend which will receive
       *                          the question
       * @param {String} question the new question
       *
       * @returns {Object} `undefined` if the user didn't provide any question,
       *                   otherwise an object with the following fields:
       *
       *                       {
       *                         "otherAnswer": "...",
       *                         "unsafe": true,
       *                         "myAnswer": "...",
       *                         "myQuestion": "..."
       *                       }
       *
       *                   The last two fields are available only if `unsafe` is
       *                   true.
       */
      Snake.Crypto.SMP.answerAndAskBack = function (username, question) {
        return Modal.enqueue("Friend authentication",
                             Snake.instanceTemplate("friend-answer-modal",
                                                    {
                                                        "friend-name": username,
                                                        "question": question
                                                    })
                            );
      };

    }
  });

}());
