(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View used to manage all the groups of an user.
     * @class Snake.View.ManageGroups
     */
    ManageGroups: Snake.View.ViewBase.extend({
      nestedViews: ["form", "groups"],
      events: {
        "click #add-group": "addNew"
      },
      initialize: function(options) {
        this.groups = new Snake.View.UpdatingCollection({
          collection: this.model.groups,
          renderFilter: function(group) {
            return group.isStandalone();
          },
          childViewConstructor: Snake.View.ListGroups,
          childViewOptions: {template: "my-groups"}
        });

        this.form = new Snake.View.Form({
          schema: Snake.Schema.group,
          legend: "Your group",
          button: "Save changes",
        });

        var groups = this.model.groups;
        groups.bind("item:leave", this.leave, this);
        // XXX this is awful
        groups.bind("item:open", function(args) {
          window.location.hash = "#group/" + args.id;
        }, this);
        // XXX this is awful too
        groups.bind("item:admin", function(args) {
          window.location.hash = "#groupadmin/" + args.id;
        }, this);

        this.form.bind("form:close", this.hideForm, this);
        this.form.bind("form:success", this.hideForm, this);
      },
      hideForm: function() {
        this.newGroupForm.hide();
        this.addGroupButton.show();
      },
      showForm: function() {
        this.newGroupForm.show();
        this.addGroupButton.hide();
      },
      leave: function(args) {
        var id = args.id;
        var group = this.model.groups.get(id);

        // Standalone groups cannot be leaved
        if (group.isStandalone()) {
          this.showAlertError("You're trying to do something bad.");
          return;
        }

        // TODO send a message to leave the group
      },
      edit: function(args) {
        var id = args.id;
        var group = this.model.groups.get(id);

        // Standalone groups cannot be changed
        if (group.isStandalone()) {
          this.showAlertError("You're trying to do something bad.");
          return;
        }

        this.form.changeModel(group);
        this.form.beforeSave = Snake.identityPromise;
      },
      addNew: function() {
        // We create a bogus group here, just to gather the user input
        var group = new Snake.Model.Group({}, {setDefaultValues: true});

        // remember that the real crypto is done in the form.submit callback
        this.form.changeModel(group);
        this.form.beforeSave = function (model, formData) {
          return model.setupNew(formData);
        };
        this.showForm();

        this.listenToOnce(group, 'sync', function() {
          return group.save().result;
        });
      },
      render: function() {
        var newManageProfile;

        newManageProfile = Snake.instanceTemplate("manage-groups",
            this.model.attributeTree(), this.el.id);

        this.renderAllNestedViews(newManageProfile);

        this.$el.replaceWith(newManageProfile);
        this.setElement(newManageProfile[0]);

        this.trigger("title", this.title());

        this.alertSuccess = this.$el.find("#manageGroupSuccess");
        this.alertError = this.$el.find("#manageGroupError");

        this.newGroupForm = this.$el.find("#newGroupForm");
        this.addGroupButton = this.$el.find("#add-group");

        return this;
      },
      title: function() {
        return "Manage your groups";
      }
    })
  });
}());
