(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for creating a new Message.
     * @class Snake.View.MessageEditor
     */
    MessageEditor: Snake.View.ViewBase.extend({
      events: {
        "submit": "submit"
      },
      initialize: function(options) {
        this.options = options || {};
        this.collection = options.collection;

        // Initialize empty post model
        this.reset();
      },
      rendered: false,
      render: function() {
        // Render once only
        if (!this.rendered) {
          this.rendered = true;

          // Create a new instance the template
          var newEditor;

          newEditor = Snake.instanceTemplate("message-editor",
              this.model.attributeTree(), this.el.id);

          this.$el.replaceWith(newEditor);
          this.setElement(newEditor[0]);
        }

        return this;
      },
      /**
       * Clears the textbox text and initialize a new Message model.
       * If a collection has been specified, adds the previous model
       * to it.
       * @memberOf Snake.View.MessageEditor#
       */
      reset: function() {
        if (!_.isUndefined(this.model) && this.model.has("_recipientId")) {
          this.options.recipient = this.model.get("_recipientId");
        }

        this.model = new Snake.Model.Message({
          _senderId: Snake.sessionUser.id,
          _recipientId: this.options.recipient,
          _recipientType: "user",
          type: "post"
        }, this.options.message);

        this.$el.find("textarea").val("");
      },
      /**
       * Fired when the post is submitted. Sets date and text and saves it.
       * @memberOf Snake.View.MessageEditor#
       * @returns {Boolean} blocks the form submission.
       */
      submit: function() {

        this.model.set("post", this.$el.find("textarea").val());
        this.model.set("date", new Date());

        // Save the message
        this.model.send().then(function() {

          this.model.fetchNestedModels();
          if (!_.isUndefined(this.collection)) {
            this.collection.add(this.model);
          }

          this.reset();
        }.bind(this));

        return false;
      },
      /**
       * Sets the ID of the recipient.
       * @memberOf Snake.View.PostEditor#
       * @param {type} id the new ID of the recipient group.
       */
      setRecipient: function(id) {
        this.model.set("_recipientId", id);
      }
    })
  });
}());
