(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * Manages the whole page, decides which view to show.
     * @class Snake.View.Page
     */
    Page: Snake.View.ViewBase.extend({
      id: "content",
      initialize: function(options) {
        options = options || {};
        this.pinnedPages = options.pinnedPages || {};

        this.pages = [];
        // Force to use addPage for initial pages
        if (!_.isUndefined(options.pages)) {
          _(_(options.pages).keys()).each(function(key) {
            this.addPage(key, options.pages[key]);
          }, this);
        }
        this.listenTo(this.model, "change:page", this.render);
      },
      render: function() {
        var page, menu;
        page = this.model.get("page");

        // Hide all the pages...
        _(_(this.pages).keys()).each(function(page) {
          this.pages[page].$el.hide();
        }, this);
        // ...and show the correct one
        this.pages[page].$el.show();

        // Associate renderTitle with the current page's title
        this.stopListening(null, null, this.renderTitle);
        this.listenTo(this.pages[page], "title", this.renderTitle);
        // Render the current page main view
        this.pages[page].render();

        return this;
      },
      /**
       * Renders the navigation menu.
       * @memberOf Snake.View.Page#
       */
      renderMenu: function() {
        var menu, currentPage, directives;
        menu = [];
        currentPage = this.model.get("page");

        // Add pinned pages
        _(_(this.pinnedPages).keys()).each(function(pinnedPage) {
          menu.push({entry: {
              page: pinnedPage,
              title: this.pinnedPages[pinnedPage]
            }});
        }.bind(this));

        // Prepare the input data for rendering, it will be used
        // exclusively from directives
        _(_(this.pages).keys()).each(function(page) {
          // Don't readd pinned pages
          if (page in this.pinnedPages) return;

          menu.push({entry: {
              page: page,
              title: this.pages[page].title()
            }});
        }.bind(this));

        // Rules to render the menu
        directives = {
          entry: {
            link: {
              href: function() {
                return "#" + this.page;
              },
              text: function() {
                return this.title;
              }
            },
            class: function() {
              // Check if it's the current page
              if (this.entry.page === currentPage) {
                return "active";
              }
            }
          }
        };

        $("#menu").render(menu, directives);
      },
      /**
       * Fired when the current page's title is updated
       * @memberOf Snake.View.Page#
       * @param {type} title the new title.
       */
      renderTitle: function(title) {
        document.title = title;
      },
      /**
       * Adds a page to the internal collection of pages, registers the
       * title event and adds its root element to the appropriate element.
       * @memberOf Snake.View.Page#
       * @param {type} page the new page's path.
       * @param {type} view the associated view.
       */
      addPage: function(page, view) {
        this.pages[page] = view;
        view.$el.appendTo(this.el);
        this.listenTo(view, "title", this.renderMenu);
      }
    })
  });
}());
