(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for a single post.
     * @class Snake.View.Post
     */
    Post: Snake.View.ViewBase.extend({
      initialize: function() {
        // Listen to the change events of the model, the author and the
        // recipient
        this.renderOn([this.model, this.model.author.actualProfile]);
        if (!_.isUndefined(this.model.recipient)) this.renderOn(
              [this.model.recipient.friendsOf.actualProfile]);
      },
      render: function() {
        var attributes, directives, newPost;

        // TODO: can we factorize this so we don't have to convert
        // non-trivial data types each time in the view, but let it
        // handl to Snake.instanceTemplate?
        attributes = this.model.attributeTree();
        attributes.date = Snake.formatDate(attributes.date);

        // Simplify
        directives = {};
        directives.sender = {
          text: function() {
            var profile = this.author.actualProfile;
            return profile.name + " " + profile.surname;
          },
          href: function() {
            return "#profile/" + this.author.actualProfile.id;
          }
        };
        if (!_.isUndefined(this.model.recipient)) {
          if (this.model.recipient.isStandalone()) {
            directives.recipient = {
              text: function() {
                return this.recipient.name;
              },
              href: function() {
                return "#group/" + this._recipientId;
              }
            };
          } else {
            directives.recipient = {
              text: function() {
                var profile = this.recipient.friendsOf.actualProfile;
                return profile.name + " " + profile.surname;
              },
              href: function() {
                return "#profile/" + this.recipient.friendsOf.id;
              }
            };
          }
        }

        newPost = Snake.instanceTemplate("post", attributes, "", directives);

        if (!_.isUndefined(this.model.recipient)) {
          newPost.find(".optional-recipient").show();
        }

        this.$el.replaceWith(newPost);
        this.setElement(newPost[0]);

        return this;
      }
    })
  });
}());
