(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for creating a new Post.
     * @class Snake.View.PostEditor
     */
    PostEditor: Snake.View.ViewBase.extend({
      events: {
        "submit": "submit"
      },
      initialize: function(options) {
        this.options = options || {};
        this.collection = options.collection;

        // Initialize empty post model
        this.reset();
      },
      rendered: false,
      render: function() {
        // Render once only
        if (!this.rendered) {
          this.rendered = true;

          // Create a new instance the template
          var newEditor;

          newEditor = Snake.instanceTemplate("post-editor",
              this.model.attributeTree(), this.el.id);

          this.$el.replaceWith(newEditor);
          this.setElement(newEditor[0]);
        }

        return this;
      },
      /**
       * Clears the textbox text and initialize a new Post model.
       * If a collection has been specified, adds the previous model to it.
       * @memberOf Snake.View.PostEditor#
       */
      reset: function() {
        if (!_.isUndefined(this.model) && this.model.has("_recipientId")) {
          this.options.recipient = this.model.get("_recipientId");
        }

        this.model = new Snake.Model.Post({
          authorId: Snake.sessionUser.id,
          _recipientId: this.options.recipient,
          _recipientType: "group",
          type: "post"
        }, this.options.post);

        this.$el.find("textarea").val("");
      },
      /**
       * Fired when the post is submitted. Sets text and saves it.
       * @param {Object} event the fired event.
       * @memberOf Snake.View.PostEditor#
       */
      submit: function(event) {
        event.preventDefault();

        this.model.set("post", this.$el.find("textarea").val());
        this.model.save({}).result.then(
          function() {
            this.model.fetchNestedModels();
            if (!_.isUndefined(this.collection)) {
              this.collection.add(this.model);
            }

            this.reset();
          }.bind(this),
          function(error) {
            // TODO better error handling
            alert(error);
         }.bind(this));
      },
      /**
       * Sets the ID of the recipient.
       * @memberOf Snake.View.PostEditor#
       * @param {type} id the new ID of the recipient group.
       */
      setRecipient: function(id) {
        this.model.set("_recipientId", id);
      }
    })
  });
}());
