(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for the user profile.
     * @class Snake.View.Profile
     */
    Profile: Snake.View.ViewBase.extend({
      nestedViews: ["posts", "friends", "groups", "editor"],
      events: {
        "click .unfriend": "unfriend",
        "click .become-friend": "becomeFriend"
      },
      initialize: function(options) {
        // TODO: factorize
        this.friends = new Snake.View.UpdatingCollection({
          collection: this.model.friendsGroup.members,
          childViewConstructor: Snake.View.CompactUser,
          renderFilter: function(model) {
            return model.id !== this.model.owner.id;
          }.bind(this)
        });
        this.posts = new Snake.View.UpdatingCollection({
          collection: this.model.friendsGroup.messages,
          childViewConstructor: Snake.View.Post
        });
        this.groups = new Snake.View.UpdatingCollection({
          collection: this.model.groups,
          childViewConstructor: Snake.View.CompactGroup,
          renderFilter: function(model) {
            return !model.has("friendsOf");
          }
        });

        this.editor = new Snake.View.PostEditor({
          collection: this.posts.collection,
          post: {
            fetchAuthor: true,
            author: {
              fetchActualProfile: true
            }
          }
        });
        // When the ID of friendsGroups is available set it in the post editor
        this.listenTo(this.model.friendsGroup, "change", function() {
          this.editor.setRecipient(this.model.friendsGroup.id);
        }.bind(this));

        // Listen to the change events of the model and the profile
        this.renderOn([this.model, this.model.owner]);
      },
      render: function() {
        // Create a new instance the template
        var newProfile;

        if (this.model.owner.id === Snake.sessionUser.get("id")) {
          newProfile = Snake.instanceTemplate("my-user",
              this.model.attributeTree(), this.el.id, directives);
        } else {
          var directives = {
            "messages": {
              href: function() {
                return "#messages/" + this.owner.id;
              }
            }
          };
          newProfile = Snake.instanceTemplate("user",
              this.model.attributeTree(), this.el.id, directives);
        }

        this.renderAllNestedViews(newProfile);

        this.$el.replaceWith(newProfile);
        this.setElement(newProfile[0]);

        if (Snake.sessionUser.isFriend(this.model.owner.id)) {
          newProfile.find(".optional-friends").show();
        } else {
          newProfile.find(".optional-not-friends").show();
        }

        this.trigger("title", this.title());

        return this;
      },
      title: function() {
        var profile = this.model;
        return profile.get("name") + " " + profile.get("surname") +
            "'s profile";
      },
      unfriend: function() {
        Snake.sessionUser.unfriend(this.model.owner.id);
        this.render();
      },
      becomeFriend: function() {
        Snake.sessionUser.becomeFriend(this.model.owner.id);
        this.render();
      }
    })
  });

}());
