(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View for the user's stream.
     * @class Snake.View.Stream
     */
    Stream: Snake.View.ViewBase.extend({
      nestedViews: ["messages", "editor"],
      initialize: function(options) {
        this.messageCollection = new Snake.Model.Collection([], {
          model: Snake.Model.Post,
          comparator: Snake.dateComparator()
        });

        this.messages = new Snake.View.UpdatingCollection({
          collection: this.messageCollection,
          childViewConstructor: Snake.View.Post
        });

        // TODO profiles[0] is a public profile (which does not have a
        //      friendsGroup)
        // TODO allow choice of the profile where to post
        var sessionUserProfile = Snake.sessionUser.profiles.at(1);

        this.editor = new Snake.View.PostEditor({
          recipient: sessionUserProfile.friendsGroup.id,
          collection: this.messageCollection,
          post: {
            fetchAuthor: true,
            fetchRecipient: true,
            recipient: {
              fetchFriendsOf: true
            },
            author: {
              fetchActualProfile: true
            }
          }
        });

        // TODO is this necessary?
        // When the ID of friendsGroups is available set it in the post editor
        sessionUserProfile.friendsGroup.on("change:id", function() {
          this.editor.setRecipient(sessionUserProfile.friendsGroup.id);
        }, this);

        this.model.groups.on("add", this.groupAdd, this);
        this.model.groups.on("remove", this.groupRemove, this);
      },
      groupAdd: function(model, collection, options) {
        model.messages.on("add", this.messageAdded, this);
        model.messages.on("remove", this.messageRemoved, this);
      },
      groupRemove: function(model, collection, options) {
        model.messages.off("add", this.messageAdded, this);
        model.messages.off("remove", this.messageRemoved, this);
      },
      messageAdded: function(model, collection, options) {
        this.messageCollection.add([model]);
      },
      messageRemoved: function(model, collection, options) {
        this.messageCollection.remove([model]);
      },
      render: function() {
        // Create a new instance the template
        var newStream;

        newStream = Snake.instanceTemplate("stream",
            this.model.attributeTree(), this.el.id);

        this.renderAllNestedViews(newStream);

        this.$el.replaceWith(newStream);
        this.setElement(newStream[0]);

        this.trigger("title", this.title());

        return this;
      },
      title: function() {
        return "Stream";
      }
    })
  });
}());
