(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View associated to a Underscore's collection, automatically update
     * when an item is added or removed.
     * @class Snake.View.UpdatingCollection
     * @link http://liquidmedia.org/blog/2011/02/backbone-js-part-3/
     */
    UpdatingCollection: Backbone.View.extend({
      initialize: function(options) {
        _(this).bindAll('add', 'remove');

        if (_.isUndefined(options.childViewConstructor)) {
          throw new Error("No child view constructor provided");
        }

        this.renderFilter = options.renderFilter;
        this._childViewConstructor = options.childViewConstructor;
        this._childViewOptions = options.childViewOptions || {};
        // this._childViewTagName = options.childViewTagName;

        this._childViews = [];

        this.collection.each(this.add);

        this.collection.bind('add', this.add);
        this.collection.bind('remove', this.remove);

        this.collection.bind('sort', function() {
          this._childViews = [];
          this.collection.each(this.add);
          this.render();
        }.bind(this));
      },
      add: function(model) {
        var childView, childViewOptions, index, insertIndex;

        insertIndex = index = this.collection.indexOf(model);

        if (_.isUndefined(this.renderFilter) || this.renderFilter(model)) {

          childViewOptions = _.extend(this._childViewOptions, {
            model: model
            // tagName: this._childViewTagName,
          });

          childView = new this._childViewConstructor(childViewOptions);

          if (this.rendered) {
            // Look for rendered view in _childViews or wait for
            // the end of the underlying collection
            while (index < this._childViews.length &&
                _.isUndefined(this._childViews[index])) {
              index++;
            }

            // Insert the new element in the right place
            if (index >= this._childViews.length) {
              this.$el.append(childView.render().el);
            } else {
              childView.render().$el.insertBefore(this._childViews[index].el);
            }
          }
        }

        // Insert new view in childViews or undefined if we skipped it
        // We need this check because if insertIndex is greater than
        // array's length doesn't insert at the specified position
        // but simply appends
        if (insertIndex >= this._childViews.length) {
          this._childViews[insertIndex] = childView;
        } else {
          this._childViews.splice(insertIndex, 0, childView);
        }

      },
      remove: function(model) {
        var viewToRemove = _(this._childViews).select(function(cv) {
          return cv.model === model;
        })[0];
        this._childViews = _(this._childViews).without(viewToRemove);

        if (this.rendered) {
          $(viewToRemove.el).remove();
        }
      },
      render: function() {
        this.rendered = true;

        $(this.el).empty();

        _(this._childViews).each(function(childView) {
          if (!_(childView).isUndefined()) {
            $(this.el).append(childView.render().el);
          }
        }, this);

        return this;
      }
    })
  });
}());
