// DOM Ready
$(function() {
  "use strict";

  function getQueryVariable(name) {
    var query = window.location.search.substring(1);
    var vars = query.split("&");
    for (var i = 0; i < vars.length; i++) {
      var pair = vars[i].split("=");
      if(pair[0] === name)
        return pair[1];
    }
    return "";
  }

  function setManualAuthenticatorResponse(response) {
      Snake.Crypto.ManualAuthenticator.askUser = function(name, pubKey) {
        return Snake.identityPromise(response);
      };
  }

  Snake.View = _.extend(Snake.View || {}, {
    init: function() {
      // Always confirm public keys
      setManualAuthenticatorResponse(true);

      // Show all profiles
      Snake.Model.SessionUser.chooseProfiles = function (profileData) {
        return Snake.identityPromise(_.pluck(profileData, "id"));
      };
    }
  });

  function pw(test) {
    return function (done) {
      return test.bind(this)().then(done.bind(null, null), done);
    };
  }

  /**
   * Register one or more users with the specified usernames, using as password
   * the username itself. If a user with that name is already registered, it
   * does nothing.
   *
   * @param {String[]} usernames array of usernames to register.
   * @returns {Promise} a Promise that will resolve when the registration of all
   * the user is terminated.
   */
  function registerUsers(usernames) {
    var result = Snake.identityPromise();

    _.forEach(usernames, function (username) {
      var user, register;

      result = result.then(function () {
        if (config.networkBackend === "remote") {
          return Snake.IO.read({url: {
            object: "username",
            username: username
          }}, {rawResponse: true}).result.catch(function () {
            return Model.SessionUser.register(username, username);
          });
        } else {
          user = _.findByField(Snake.tables.users.data, username, "username");
          if (_.isUndefined(user)) {
            return Model.SessionUser.register(username, username);
          }
        }
      });
    });

    return result;
  }

  var sessionUsers = [];
  /**
   * Setups users and puts them in the sessionUsers array
   *
   * @param {Array} names: array of names of the users
   * @returns {Function} a function that returns a result Promise
   */
  function setupUsers(names) {
    var result, login;
    return function() {
      Snake.IO.Cache.flush();

      result = registerUsers(names);
      sessionUsers = [];

      _.each(names, function(el, i){
        login = Model.SessionUser.login.bind(null, el, el);
        result = result.then(login).then(function () {
          sessionUsers[i] = Snake.sessionUser;
          Snake.IO.Cache.flush();
          Snake.sessionUser = null;
        });
      });

      return result.then(function () {
        Snake.sessionUser = null;
      });
    };
  }

  function quickFriendship(indexA, indexB) {
    var userA, userB;

    userA = sessionUsers[indexA];
    userB = sessionUsers[indexB];

    setManualAuthenticatorResponse(true);
    Snake.sessionUser = userA;
    return userA.becomeFriend(userB.id).then(function () {
      Snake.sessionUser = userB;
      return Snake.sessionUser.processNewMessages();
    }).then(function () {
      Snake.sessionUser = userA;
      return Snake.sessionUser.processNewMessages();
    }).then(function () {
      Snake.sessionUser = userB;
      return Snake.sessionUser.processNewMessages();
    }).then(function () {
      Snake.sessionUser = userA;
      return Snake.sessionUser.processNewMessages();
    });
  }

  var timers = {};
  var perf = window.performance;

  function Timer(name) {
    var startTime = perf.now();
    return function (forward) {
      var now = perf.now();
      timers[name] = now - startTime;
      startTime = now;
      return Snake.identityPromise(forward);
    };
  }

  // Display configuration
  var config = {
    networkBackend: getQueryVariable("network") === "remote" ? "remote" : "local",
    cryptoBackend: getQueryVariable("crypto") === "fake" ? "fake" : "real",
    repeat: _.isUndefined(getQueryVariable("repeat")) ? Infinity : new Number(getQueryVariable("repeat"))
  };
  document.getElementById("network-backend").textContent = config.networkBackend;
  document.getElementById("crypto-backend").textContent = config.cryptoBackend;

  _(document.getElementsByClassName("change-config")).each(function (element) {
    var change = element.getAttribute("href").split("=");
    var newUrl = window.location.pathname + "?";
    var query = window.location.search.substring(1);
    var vars = query.split("&");
    var found = false;
    var i;
    var pair;
    for (i = 0; i < vars.length; i++) {
      if (vars[i].indexOf("=") !== -1) {
        pair = vars[i].split("=");
        newUrl += pair[0] + "=";
        if(pair[0] === change[0]) {
          found = true;
          newUrl += change[1];
        } else {
          newUrl += pair[1];
        }
        newUrl += "&";
      }
    }

    if (!found)
      newUrl += change[0] + "=" + change[1] + "&";

    element.href = newUrl.substring(0, newUrl.length - 1);
  });

  document.getElementById("clear-stats").onclick = localStorage.clear.bind(localStorage);
  document.getElementById("export-stats").onclick = function () {
    var itemName, currentData, i, prefix, result, labels, csv;
    prefix = "stats-";
    result = {};
    labels = [];

    var avg = function (data) {
      return _.reduce(data, function (acc, value) {
        return acc + value / data.length;
      }, 0);
    };

    var stddev = function (avg, data) {
      return Math.sqrt(_.reduce(data, function (acc, value) {
        return acc + Math.pow(value - avg, 2) / data.length;
      }, 0));
    };

    for (i = 0; i < localStorage.length; i++) {
      itemName = localStorage.key(i);
      if (itemName.substring(0, prefix.length) === prefix) {
        labels.push(itemName + "_avg");
        labels.push(itemName + "_stddev");
        currentData = JSON.parse(localStorage.getItem(itemName) || "[]");
        _.keys(currentData).forEach(function (key) {
          var avg_value, stddev_value;
          avg_value = avg(currentData[key]);
          stddev_value = stddev(avg_value, currentData[key]);
          result[key] = (result[key] || []).concat([avg_value, stddev_value]);
        });
      }
    }

    csv = "name," + labels.join(",") + "\n";
    _.keys(result).forEach(function(testName) {
      csv += "\"" + testName + "\"," + result[testName].join(",") + "\n";
    });

    console.log(csv);
  };

  // Initialize components
  Snake.Config.init();
  Snake.Schema.init();
  if (config.networkBackend === "remote")
    Snake.Network.init();
  else
    Snake.LocalNetwork.init();
  Snake.IO.init();
  Snake.Modal.init();
  Snake.Crypto.init();
  if (config.cryptoBackend === "fake")
    Snake.Crypto.FakeCommon.init();
  Snake.View.init();

  Snake.Network.reset({"object": "clean"});

  var Model = Snake.Model;

  mocha.setup("bdd");

  var expect = chai.expect;
  chai.use(chaiAsPromised);

  describe("Registration and login", function() {
    this.timeout(60 * 1000);

    beforeEach(pw(function () {
      Snake.IO.Cache.flush();
      Snake.sessionUser = null;
      return Snake.identityPromise();
    }));

    it("should register two users", pw(function() {
      var timer = Timer(this.test.title);
      return registerUsers(["aaa", "bbb"]).then(timer);
    }));

    it("should login with correct username and password", pw(function() {
      var timer = Timer(this.test.title);
      return registerUsers(["aaa"]).then(timer)
          .then(Model.SessionUser.login.bind(null, "aaa", "aaa")).then(timer);
    }));

    // it("should fail login due to wrong credentials", pw(function() {
    //   var register, login;

    //   register = registerUsers(["aaa"]);
    //   login = Model.SessionUser.login.bind(null, "aaa", "lolnope");

    //   return expect(register.then(login)).to.be.rejected;
    // }));
  });

  describe("Groups", function() {
    var testGroupId;

    this.timeout(60 * 1000);

    before(pw(function () {
      return setupUsers(["GroupOwner", "Friend"])().then(function () {
        return quickFriendship(0, 1);
      });
    }));

    it("should check if every user is admin of a group", pw(function () {
      var groupId, groupModel, groups;
      var timer = Timer(this.test.title);

      Snake.sessionUser = sessionUsers[0];
      groups = _.filter(Snake.sessionUser.get("groups"), function (group) {
        return _.has(group, "adminKey");
      });
      expect(groups.length === 1)

      groupId = groups[0].id;
      groupModel = new Snake.Model.Group({id: groupId}, {});
      return groupModel.fetch().result
            .then(timer)
            .then(function() {
        expect(!groupModel.isStandalone()).to.be.true;
      });
    }));

    it("should create a new group, and insert the user", pw(function () {
      var group, groupData, groupFetch;
      var timer = Timer(this.test.title);

      Snake.sessionUser = sessionUsers[0];
      group = new Snake.Model.Group({}, {setDefaultValues: true});
      groupData = {name: "test", description: "test"};
      return group.setupNew(groupData).then(timer).then(function() {
        var groups;

        groups = _.filter(Snake.sessionUser.get("groups"), function (group) {
          return _.has(group, "adminKey");
        });
        expect(groups.length === 2).to.be.true;
        expect(_.pluck(groups, "id").indexOf(group.id) !== -1);

        groupFetch = new Snake.Model.Group({id: group.id}, {
          fetchMembers: true,
          fetchAdministrators: true
        });

        return groupFetch.fetchAll();
      }).then(function () {
        var admins = groupFetch.administrators;
        // The new group should have only one administrator, the current user
        expect(groupFetch.members.length === 0);
        expect(admins.length === 1);
        expect(admins.models[0].id === Snake.sessionUser.get("id"));

        testGroupId = group.id;
      });
    }));

    it("should insert a friend in the group", pw(function () {
      var group, newMemberId, newMemberKey;
      var timer = new Timer(this.test.title);

      return Snake.identityPromise().then(function () {
        Snake.sessionUser = sessionUsers[0];
        newMemberId = sessionUsers[1].id;
        newMemberKey = sessionUsers[1].get("staticPublicKey");

        group = new Snake.Model.Group({id: testGroupId}, {
          fetchAdminData: true
        });

        return group.fetchAll();
      }).then(timer).then(function () {
        return group.insertMember(newMemberId, newMemberKey);
      }).then(timer).then(function () {
        Snake.IO.Cache.flush();
        group = new Snake.Model.Group({id: testGroupId}, {
          fetchAdministrators: true,
          fetchMembers: true
        });
        return group.fetchAll();
      }).then(function () {
        expect(group.administrators.length === 1);
        expect(group.members.length === 1);
        expect(group.members.models[0].id === newMemberId);
      });
    }));

    it("should remove the friend from the group", pw(function () {
      var group, memberId;
      var timer = Timer(this.test.title);

      return Snake.identityPromise().then(function () {
        Snake.sessionUser = sessionUsers[0];
        memberId = sessionUsers[1].id;

        group = new Snake.Model.Group({id: testGroupId}, {
          fetchAdminData: true
        });

        return group.fetchAll();
      }).then(timer).then(function () {
        return group.removeMember(memberId);
      }).then(timer).then(function () {
        Snake.IO.Cache.flush();
        group = new Snake.Model.Group({id: testGroupId}, {
          fetchAdministrators: true,
          fetchMembers: true
        });
        return group.fetchAll();
      }).then(function () {
        expect(group.administrators.length === 1);
        expect(group.members.length === 0);
      });
    }));

  });

  describe("Friendship", function () {
    this.timeout(60 * 1000);

    before(pw(setupUsers(["Bob", "Charlie", "Dan"])));

    it("create a new friendship using the manual auth", pw(function () {
      var timer = Timer(this.test.title);

      return quickFriendship(0, 1).then(timer).then(function () {
        expect(sessionUsers[1].getFriend(sessionUsers[0].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
        expect(sessionUsers[0].getFriend(sessionUsers[1].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
      });
    }));

    it("should create a new friendship using SMP", pw(function () {
      var timer = Timer(this.test.title);

      setManualAuthenticatorResponse(false);
      Snake.sessionUser = sessionUsers[0];

      Snake.Crypto.SMP.askQuestion = function () {
        return Snake.identityPromise({
            "myQuestion": "Do you like pizza?",
            "myAnswer": "42"
        });
      };

      return sessionUsers[0].becomeFriend(sessionUsers[2].id).then(function () {
        Snake.sessionUser = sessionUsers[2];

        Snake.Crypto.SMP.answerAndAskBack = function () {
          return Snake.identityPromise({
              "otherAnswer": "42",
              "unsafe": true,
              "myQuestion": "Do you like pasta?",
              "myAnswer": "lolnope"
          });
        };

        return Snake.sessionUser.processNewMessages();
      }).then(function () {
        Snake.sessionUser = sessionUsers[0];

        Snake.Crypto.SMP.answerAndAskBack = function () {
          return Snake.identityPromise({
              "otherAnswer": "lolnope",
              "unsafe": false
          });
        };

        return Snake.sessionUser.processNewMessages();
      }).then(function () {
        Snake.sessionUser = sessionUsers[2];
        return Snake.sessionUser.processNewMessages();
      }).then(function () {
        Snake.sessionUser = sessionUsers[0];
        return Snake.sessionUser.processNewMessages();
      }).then(timer).then(function () {
        expect(sessionUsers[2].getFriend(sessionUsers[0].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
        expect(sessionUsers[0].getFriend(sessionUsers[2].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
      });
    }));

    it("should send a message", pw(function () {
      var message;
      var timer = Timer(this.test.title);

      Snake.sessionUser = sessionUsers[0];

      message = new Snake.Model.Message({}, {});
      message.set("_recipientType", "user");
      message.set("_recipientId", sessionUsers[1].id);
      message.set("_senderId", sessionUsers[0].id);
      message.set("post", "Hello!");
      message.set("type", "post");
      message.set("date", new Date());

      return message.send().then(timer);
    }));

    it("read the sent message", pw(function () {
      var message;
      var timer = Timer(this.test.title);

      Snake.sessionUser = sessionUsers[1];
      return Snake.sessionUser.processNewMessages()
            .then(timer)
            .then(function (newMessages) {
        expect(newMessages.length).to.be.equal(1);
        expect(newMessages[0].get("post")).to.be.equal("Hello!");
      });
    }));

    it("read the latest messages", pw(function () {
      var messages;
      var timer = Timer(this.test.title);

      Snake.sessionUser = sessionUsers[1];

      messages = new Snake.Model.Collection([], {
        model: Snake.Model.Message,
        comparator: Snake.dateComparator(),
        url: function () {
          var friend = Snake.sessionUser.getFriend(sessionUsers[0].id);
          return {
            object: "last-messages",
            addresses: [friend.myAddress, friend.otherAddress],
            num: 10
          };
        }.bind(this)
      });

      return messages.fetch().result.then(timer).then(function () {
        var posts;

        posts = _.filter(messages.models, function (model) {
          return model.get("type") === "post";
        });

        expect(posts.length).to.be.equal(1);
        expect(posts[0].get("post")).to.be.equal("Hello!");
      });
    }));
  });

  describe("Web of Trust", function() {
    this.timeout(60 * 1000);

    before(pw(function () {
      var fail = function () {
        throw new Error("We got to SMP.");
      };
      Snake.Crypto.SMP.askQuestion = fail;
      Snake.Crypto.SMP.answerAndAskBack = fail;

      return setupUsers(["Aaaa", "Bbbb", "Cccc", "Dddd", "Eeee", "Ffff"])();
    }));

    it("should create a friendship using Web of Trust if there are 3 friends" +
      " in common", pw(function () {
      return quickFriendship(0, 1).then(function(){
        return quickFriendship(0, 2);
      }).then(function(){
        return quickFriendship(0, 3);
      }).then(function(){
        return quickFriendship(1, 4);
      }).then(function(){
        return quickFriendship(2, 4);
      }).then(function(){
        return quickFriendship(3, 4);
      }).then(function(){
         // now that 0 has 3 friends, we use the WoT to authenticate 4
         setManualAuthenticatorResponse(false);

         Snake.sessionUser = sessionUsers[0];
         return Snake.sessionUser.becomeFriend(sessionUsers[4].id)
         .then(function () {
           Snake.sessionUser = sessionUsers[4];
           return Snake.sessionUser.processNewMessages();
         }).then(function () {
           Snake.sessionUser = sessionUsers[0];
           return Snake.sessionUser.processNewMessages();
         }).then(function () {
           Snake.sessionUser = sessionUsers[4];
           return Snake.sessionUser.processNewMessages();
         }).then(function () {
           Snake.sessionUser = sessionUsers[0];
           return Snake.sessionUser.processNewMessages();
         });
      }).then(function () {
        expect(sessionUsers[4].getFriend(sessionUsers[0].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
        expect(sessionUsers[0].getFriend(sessionUsers[4].id).friendshipProgress ===
            Snake.Model.SessionUser.FriendshipProgress.Complete).to.be.true;
      });
    }));

    it("should not create a friendship using Web of Trust if only 2 friends" +
      " in common", pw(function () {
      return quickFriendship(5, 4).then(function(){
        return quickFriendship(5, 0);
      }).then(function(){
         // now that 5 has 2 friends, we try to use the WoT to authenticate 1
         setManualAuthenticatorResponse(false);

         Snake.sessionUser = sessionUsers[5];
         return Snake.sessionUser.becomeFriend(sessionUsers[1].id)
         .then(function () {
           Snake.sessionUser = sessionUsers[1];
           return Snake.sessionUser.processNewMessages();
         }).then(function () {
           Snake.sessionUser = sessionUsers[5];
           return Snake.sessionUser.processNewMessages();
         }).then(function () {
           Snake.sessionUser = sessionUsers[1];
           return Snake.sessionUser.processNewMessages();
           // this catch is here to avoid the "shared secrets don't match"
           // message of SMP failing
         }).catch(function(){}).then(function () {
           Snake.sessionUser = sessionUsers[5];
           return Snake.sessionUser.processNewMessages();
         });
      }).then(function () {
        expect(sessionUsers[1].isFriend(sessionUsers[5].id)).to.be.false;
        expect(sessionUsers[5].isFriend(sessionUsers[1].id)).to.be.false;
      });
    }));
  });

  after(function () {
    var itemName, currentData;
    itemName = "stats-" + config.networkBackend + "-" + config.cryptoBackend;
    currentData = JSON.parse(localStorage.getItem(itemName) || "{}");
    _.keys(timers).forEach(function (key) {
      currentData[key] = (currentData[key] || []).concat([timers[key]]);
    });
    localStorage.setItem(itemName, JSON.stringify(currentData));

    var len = function (x) { return x.length; };
    if (_.min(_.map(_.values(currentData), len)) < config.repeat) {
      document.location.reload(false);
    }
  });

  // HACK
  setTimeout(function () { mocha.run(); }, 1000);

});
